﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
using System.Xml;
using Centido.Core.Dto;
using Microsoft.VisualStudio.TestTools.UnitTesting;

namespace Centido.Core.Test
{
    /// <summary>
    ///   Testing the data transfer object for a reference column.
    /// </summary>
    [TestClass]
    public class RefColumnDtoTest : GenericDtoTest
    {
        /// <summary>
        ///   The current domain name.
        /// </summary>
        public const string DomainName = "FullAdminRights";

        /// <summary>
        ///   The current domain.
        /// </summary>
        private Domain _domain;

        /// <summary>
        ///   The table.
        /// </summary>
        private Table _table;

        /// <summary>
        ///   The reference column.
        /// </summary>
        private RefColumn _refColumn;

        /// <summary>
        /// Test initialiaze
        /// </summary>
        [TestInitialize]
        public void Initialize()
        {
            _domain = Domain.Build(DomainConfiguration.Load(DomainName));

            _domain.DeleteAll();

            _table = _domain.AddTable(StringProperty.RandomIdentifier());

            _refColumn = RefColumn.Add(_table, StringProperty.RandomIdentifier());
        }

        /// <summary>
        /// Test clean up
        /// </summary>
        [TestCleanup]
        public void CleanUp()
        {
            _domain.DeleteAll();
        }


        /// <summary>
        ///   Retrieves the name of the data transfer object xml tag.
        /// </summary>
        /// <returns></returns>
        protected override string GetXmlTag()
        {
            return RefColumnDto.XmlTag;
        }

        /// <summary>
        ///   Retrives the current xml schema version.
        /// </summary>
        /// <returns></returns>
        protected override string GetCurrentXmlSchemaVersion()
        {
            return RefColumnDto.CurrentXmlSchemaVersion;
        }

        /// <summary>
        ///   Creates a new data transfer object.
        /// </summary>
        /// <returns></returns>
        protected override DataTransferObject CreateDto()
        {
            return new RefColumnDto
            {
                Name = StringProperty.RandomIdentifier().ToLower(),
                DisplayName = StringProperty.StandardValue('a'),
                Description = StringProperty.ExtendedValue('a'),
                OrderIndex = 15,
                IsNotNull = true,
                DefaultValue = Guid.NewGuid(),
                HasReferenceChecking = true
            };
        }

        /// <summary>
        ///   Retrieves the list of data transfer object fields.
        /// </summary>
        /// <returns></returns>
        protected override IEnumerable<DtoFieldInfo> GetFields()
        {
            yield return new DtoFieldInfo
            {
                Name = "Name",
                XmlName = ColumnDto.NameAttribute,
                IsRequired = true,
                FieldType = DtoFieldType.String,
                SampleValue = StringProperty.RandomIdentifier()
            };
            yield return new DtoFieldInfo
            {
                Name = "DisplayName",
                XmlName = ColumnDto.DisplayNameAttribute,
                IsRequired = false,
                FieldType = DtoFieldType.String,
                SampleValue = StringProperty.StandardValue('a')
            };
            yield return new DtoFieldInfo
            {
                Name = "Description",
                XmlName = ColumnDto.DescriptionAttribute,
                IsRequired = false,
                FieldType = DtoFieldType.String,
                SampleValue = StringProperty.ExtendedValue('a')
            };
            yield return new DtoFieldInfo
            {
                Name = "OrderIndex",
                XmlName = ColumnDto.OrderIndexAttribute,
                IsRequired = false,
                FieldType = DtoFieldType.Integer,
                SampleValue = 7
            };
            yield return new DtoFieldInfo
            {
                Name = "IsNotNull",
                XmlName = ColumnDto.IsNotNullAttribute,
                IsRequired = false,
                FieldType = DtoFieldType.Boolean,
                SampleValue = true
            };
            yield return new DtoFieldInfo
            {
                Name = "DefaultValue",
                XmlName = RefColumnDto.DefaultValueAttribute,
                IsRequired = false,
                FieldType = DtoFieldType.Guid,
                SampleValue = Guid.NewGuid()
            };
            yield return new DtoFieldInfo
            {
                Name = "HasReferenceChecking",
                XmlName = RefColumnDto.HasReferenceCheckingAttribute,
                IsRequired = false,
                FieldType = DtoFieldType.Boolean,
                SampleValue = true
            };
        }


        /// <summary>
        ///   Copies the properties of the data transfer object to the corresponding object.
        /// </summary>
        /// <param name="dto">The data transfer object.</param>
        /// <returns>Returns the corresponding object after copying.</returns>
        protected override object CopyFrom(DataTransferObject dto)
        {
            ((RefColumnDto)dto).CopyTo(_refColumn);

            return _refColumn;
        }


        /// <summary>
        ///   Creates a data transfer object from the correpsonding object.
        /// </summary>
        /// <param name="dto">The created data transfer object.</param>
        /// <param name="initialObject">The initial object.</param>
        protected override void CreateDtoFromObject(out DataTransferObject dto, out object initialObject)
        {
            dto = RefColumnDto.CreateFrom(_refColumn);
            initialObject = _refColumn;
        }


        /// <summary>
        ///   Reading the reference column target list from xml.
        /// </summary>
        [TestMethod]
        public void ReadingRefColumnTargetListFromXml()
        {
            // Creating the enum column
            var xml = string.Format("<{0} {1}=\"{2}\">",
                RefColumnDto.XmlTag,
                ColumnDto.NameAttribute,
                StringProperty.RandomIdentifier());

            // Creating the enumeration member list
            xml += string.Format("<{0}><{1} {2}=\"{3}\" /><{4} {5}=\"{6}\" /></{7}></{8}>",
                RefColumnDto.TargetListTag,
                RefTargetDto.XmlTag,
                RefTargetDto.RefTableNameAttribute,
                Guid.NewGuid(),
                RefTargetDto.XmlTag,
                RefTargetDto.RefTableNameAttribute,
                Guid.NewGuid(),
                RefColumnDto.TargetListTag,
                RefColumnDto.XmlTag);

            var stream = new XmlTextReader(new MemoryStream(Encoding.UTF8.GetBytes(xml)));

            var refColumnDto = new RefColumnDto();

            refColumnDto.ReadXml(stream);

            Assert.AreEqual(refColumnDto.Targets.Count(), 2);
        }


        /// <summary>
        ///   Reading an empty reference column target list from xml.
        /// </summary>
        [TestMethod]
        public void ReadingEmptyRefColumnTargetListFromXml()
        {
            // Creating the enum column
            var xml = string.Format("<{0} {1}=\"{2}\">",
                RefColumnDto.XmlTag,
                ColumnDto.NameAttribute,
                StringProperty.RandomIdentifier());

            // Creating the target member list
            xml += string.Format("<{0} /></{1}>",
                RefColumnDto.TargetListTag,
                RefColumnDto.XmlTag);

            var stream = new XmlTextReader(new MemoryStream(Encoding.UTF8.GetBytes(xml)));

            var refColumnDto = new RefColumnDto();

            refColumnDto.ReadXml(stream);

            Assert.AreEqual(refColumnDto.Targets.Count(), 0);
        }


        /// <summary>
        ///   Writing and reading the reference column target list.
        /// </summary>
        [TestMethod]
        public void WritingAndReadingTargetList()
        {
            var refColumnDto = new RefColumnDto
            {
                Name = "ref_column"
            };

            refColumnDto.AddTarget(new RefTargetDto
            {
                RefTableName = "table_1"
            });
            refColumnDto.AddTarget(new RefTargetDto
            {
                RefTableName = "table_2",
                HasCascadeDelete = true
            });

            var memoryStream = new MemoryStream();
            var writer = new XmlTextWriter(memoryStream, Encoding.UTF8);

            refColumnDto.WriteXml(writer);

            writer.Flush();

            memoryStream.Position = 0;
            var streamReader = new StreamReader(memoryStream);
            var xml = streamReader.ReadToEnd();

            Assert.IsFalse(string.IsNullOrEmpty(xml));

            memoryStream.Position = 0;
            var reader = new XmlTextReader(memoryStream);

            var readRefColumnDto = new RefColumnDto();

            readRefColumnDto.ReadXml(reader);

            Assert.AreEqual(readRefColumnDto.Targets.Count(), 2);
        }
    }
}
