﻿// -----------------------------------------------------------------------
// <copyright file="EnumMemberDto.cs" company="JSC Centido">
//  Copyright (c) JSC Centido. All rights reserved.
// </copyright>
// -----------------------------------------------------------------------

using System;
using System.Globalization;
using System.Xml;
using System.Xml.Schema;

namespace Centido.Core.Dto
{
    /// <summary>
    ///   A data transfer object for an enumeration member.
    /// </summary>
    public class EnumMemberDto : DataTransferObject
    {
        /// <summary>
        ///   The name of the xml tag that is used to store this object.
        /// </summary>
        public const string XmlTag = "enumMember";

        /// <summary>
        ///   The current xml schema version.
        /// </summary>
        public const string CurrentXmlSchemaVersion = "1";

        /// <summary>
        ///   The name of the xml attribute that is used to store an enumeration member value.
        /// </summary>
        public const string ValueAttribute = "value";

        /// <summary>
        ///   The name of the xml attribute that is used to store an enumeration member display name.
        /// </summary>
        public const string DisplayNameAttribute = "displayName";


        /// <summary>
        ///   Gets or sets the value of the enumeration member.
        /// </summary>
        public int Value { get; set; }

        /// <summary>
        ///   Gets or sets the display name of the enumeration member.
        /// </summary>
        public string DisplayName { get; set; }


        /// <summary>
        ///   This method is reserved and should not be used. When implementing the IXmlSerializable interface, you should return null (Nothing in Visual Basic) from this method, and instead, if specifying a custom schema is required, apply the <see cref="T:System.Xml.Serialization.XmlSchemaProviderAttribute"/> to the class.
        /// </summary>
        /// <returns>
        ///   An <see cref="T:System.Xml.Schema.XmlSchema"/> that describes the XML representation of the object that is produced by the <see cref="M:System.Xml.Serialization.IXmlSerializable.WriteXml(System.Xml.XmlWriter)"/> method and consumed by the <see cref="M:System.Xml.Serialization.IXmlSerializable.ReadXml(System.Xml.XmlReader)"/> method.
        /// </returns>
        public override XmlSchema GetSchema()
        {
            return null;
        }

        /// <summary>
        ///   Generates an object from its XML representation.
        /// </summary>
        /// <param name="reader">The <see cref="T:System.Xml.XmlReader"/> stream from which the object is deserialized. 
        /// </param>
        /// <exception cref="ArgumentNullException">Thrown when the value of <paramref name="reader"/> is null.</exception>
        public override void ReadXml(XmlReader reader)
        {
            if (reader == null)
                throw new ArgumentNullException("reader");

            reader.MoveToContent();

            // Reading the xml schema version
            var xmlSchemaVersion = ReadStringAttribute(reader, XmlSchemaVersionTag, isRequired: false);

            // If the xml schema version is not defined
            if (string.IsNullOrEmpty(xmlSchemaVersion)
                // or if it is the current xml schema version
                || xmlSchemaVersion == CurrentXmlSchemaVersion)
            {
                // Reading the enumeration member value
                Value = ReadIntAttribute(reader, ValueAttribute).GetValueOrDefault();

                // Reading the display name
                DisplayName = ReadStringAttribute(reader, DisplayNameAttribute, isRequired: false);

                var isEmptyElement = reader.IsEmptyElement;

                reader.ReadStartElement();

                if (!isEmptyElement) reader.ReadEndElement();
            }
            else
                throw new ArgumentException(string.Format(
                    "The specified xml schema version is not supported (node = {0}, version = {1}).",
                    reader.Name,
                    xmlSchemaVersion));
        }

        /// <summary>
        ///   Converts an object into its XML representation.
        /// </summary>
        /// <param name="writer">The <see cref="T:System.Xml.XmlWriter"/> stream to which the object is serialized. 
        /// </param>
        /// <exception cref="ArgumentNullException">Thrown when the value of <paramref name="writer"/> is null.</exception>
        public override void WriteXml(XmlWriter writer)
        {
            if (writer == null) 
                throw new ArgumentNullException("writer");

            writer.WriteStartElement(XmlTag);

            // The xml schema version
            writer.WriteAttributeString(XmlSchemaVersionTag, CurrentXmlSchemaVersion);

            // The enumeration member value
            writer.WriteAttributeString(ValueAttribute, 
                Value.ToString(CultureInfo.InvariantCulture));

            // The enumeration member display name
            if (!string.IsNullOrEmpty(DisplayName))
                writer.WriteAttributeString(DisplayNameAttribute, DisplayName);

            writer.WriteEndElement();
        }


        /// <summary>
        ///   Copies the properties of the data transfer object to the
        /// specified enumeration member.
        /// </summary>
        /// <param name="enumMember">The enumeration member.</param>
        /// <exception cref="ArgumentNullException">Thrown when the value of <paramref name="enumMember"/> is null.</exception>
        public void CopyTo(EnumMember enumMember)
        {
            if (enumMember == null)
                throw new ArgumentNullException("enumMember");

            if (Value != enumMember.Value)
                enumMember.Value = Value;

            if (DisplayName != enumMember.DisplayName)
                enumMember.DisplayName = DisplayName;
        }


        /// <summary>
        ///   Creates a data transfer object for the specified enumeration member.
        /// </summary>
        /// <param name="enumMember">The specified enumeration member.</param>
        /// <exception cref="ArgumentNullException">Thrown when the value of <paramref name="enumMember"/> is null.</exception>
        /// <returns>Returns the data transfer object for the specified enumeration member.</returns>
        public static EnumMemberDto CreateFrom(EnumMember enumMember)
        {
            if (enumMember == null)
                throw new ArgumentNullException("enumMember");

            return new EnumMemberDto
            {
                Value = enumMember.Value,
                DisplayName = enumMember.DisplayName
            };
        }
    }
}
