﻿// -----------------------------------------------------------------------
// <copyright file="DecimalColumn.cs" company="JSC Centido">
//  Copyright (c) JSC Centido. All rights reserved.
// </copyright>
// -----------------------------------------------------------------------

using System.Data.Odbc;

namespace Centido.Core
{
    using System;


    /// <summary>
    ///   A decimal table column.
    /// </summary>
    /// <example>
    ///   The following example shows how to use a decimal column.
    /// <code>
    ///     using System;
    ///     using System.Data.Common;
    ///     using Centido.Core;
    /// 
    ///     class Sample
    ///     {
    ///         static void Main()
    ///         {
    ///             var domainConfig = new DomainConfiguration()
    ///             {
    ///                 DbConnectionString = @"Driver={SQL Server};Server=.\SQLEXPRESS;Database=MyDb;Uid=admin;Pwd=P@ssw0rd;",
    ///                 DbProviderName = "MSSQLSERVER"
    ///             };
    ///         
    ///             var domain = Domain.Build(domainConfig);
    ///         
    ///             domain.DeleteAll();
    ///             
    ///             //Creating a table
    ///             var table = domain.AddTable("my_table");
    ///             table.GrantAll();
    /// 
    ///             //Creating a decimal column
    ///             var column = table.AddColumn("my_column", ColumnType.Decimal)
    ///                 as DecimalColumn;
    ///             
    ///             //Setting the column precision
    ///             column.Precision = 2;
    /// 
    ///             try
    ///             {
    ///                 //Trying to insert a value that is too large
    ///                 table.InsertRow(column.Value(100M));
    ///             }
    ///             catch(DbException)
    ///             {
    ///                 Console.WriteLine("Trying to insert a value that is too large");
    ///             }
    /// 
    ///             //Setting the column scale
    ///             column.Scale = 1;
    /// 
    ///             //Inserting a value with two digits after commma
    ///             var id = table.InsertRow(column.Value(0.25M));
    /// 
    ///             Console.WriteLine("Initial value = {0}, inserted value = {1}",
    ///                 0.25M, column.Read(id).GetDecimal());
    ///         }
    ///     }
    /// </code>
    /// </example>
    public sealed class DecimalColumn : Column, IDefaultable<decimal?>, IBoundable<decimal?>
    {
        /// <summary>
        ///   The default value of the column.
        /// </summary>
        private decimal? _defaultValue;

        /// <summary>
        ///   The maximum value of the column.
        /// </summary>
        private decimal? _maximum;

        /// <summary>
        ///   The minimum value of the column.
        /// </summary>
        private decimal? _minimum;

        /// <summary>
        ///   The precision of the column.
        /// </summary>
        private int? _precision;

        /// <summary>
        ///   The scale of the column.
        /// </summary>
        private int? _scale;

        /// <summary>
        ///   Initializes a new instance of the <see cref = "Centido.Core.DecimalColumn" /> class
        ///   for the specified column id and parent table.
        /// </summary>
        /// <param name = "columnId">The column id.</param>
        /// <param name = "table">The parent table.</param>
        /// <exception cref="ArgumentNullException">Thrown when the value of <paramref name="table"/> is null.</exception>
        internal DecimalColumn(Guid columnId, Table table)
            : base(columnId, table)
        {
        }

        /// <summary>
        ///   Initializes a new instance of the <see cref = "Centido.Core.DecimalColumn" /> class
        ///   for the specified column id, name, display name,
        /// description, nullability, order index, customization level and parent table.
        /// </summary>
        /// <param name="columnId">The column id.</param>
        /// <param name="table">The parent table.</param>
        /// <param name="columnName">The column name.</param>
        /// <param name="displayName">The column display name.</param>
        /// <param name="description">The column description.</param>
        /// <param name="isNotNull">The column ability to hold null values.</param>
        /// <param name="orderIndex">The column order index.</param>
        /// <param name="customizationlevel">The column customization level.</param>
        /// <exception cref="ArgumentNullException">Thrown when the value of <paramref name="table"/> is null.</exception>
        /// <exception cref="ArgumentNullException">Thrown when the value of <paramref name="columnName"/> is null.</exception>
        /// <exception cref="ArgumentException">Thrown when the value of <paramref name="columnName"/> is an empty string.</exception>
        internal DecimalColumn(Guid columnId, Table table,
            string columnName, string displayName,
            string description, bool isNotNull,
            int orderIndex, int customizationlevel) : base(columnId, table,
            columnName, displayName, description, isNotNull, orderIndex, customizationlevel)
        {
        }

        /// <summary>
        ///   Gets the column type.
        /// </summary>
        public override ColumnType ColumnType
        {
            get { return ColumnType.Decimal; }
        }

        /// <summary>
        ///   Gets the column value database type.
        /// </summary>
        public override OdbcType ValueOdbcType
        {
            get { return OdbcType.Decimal; }
        }

        /// <summary>
        ///   Gets or sets the precision of the column.
        /// </summary>
        /// <remarks>
        /// <para>The precision determines the number of digits that can have a value stored in the column.
        /// The precision of the column must always be greater or equal to the scale of the column.</para>
        /// <para>The current user has to have the permission to change the schema in order to set this property.</para>
        /// </remarks>
        /// <exception cref="System.Data.Common.DbException">Thrown when the current user has no permission to set this property.</exception>
        /// <exception cref="System.Data.Common.DbException">Thrown when trying to set the precision less than the scale.</exception>
        /// <example>
        ///   The following example shows how to set the precision of a decimal column.
        /// <code>
        ///     using System;
        ///     using System.Data.Common;
        ///     using Centido.Core;
        /// 
        ///     class Sample
        ///     {
        ///         static void Main()
        ///         {
        ///             var domainConfig = new DomainConfiguration()
        ///             {
        ///                 DbConnectionString = @"Driver={SQL Server};Server=.\SQLEXPRESS;Database=MyDb;Uid=admin;Pwd=P@ssw0rd;",
        ///                 DbProviderName = "MSSQLSERVER"
        ///             };
        ///         
        ///             var domain = Domain.Build(domainConfig);
        ///         
        ///             domain.DeleteAll();
        ///             
        ///             //Creating a table
        ///             var table = domain.AddTable("my_table");
        ///             table.GrantAll();
        /// 
        ///             //Creating a decimal column
        ///             var column = table.AddColumn("my_column", ColumnType.Decimal)
        ///                 as DecimalColumn;
        ///             
        ///             //Setting the column precision
        ///             column.Precision = 2;
        /// 
        ///             try
        ///             {
        ///                 //Trying to insert a value that is too large
        ///                 table.InsertRow(column.Value(100M));
        ///             }
        ///             catch(DbException)
        ///             {
        ///                 Console.WriteLine("Trying to insert a value that is too large");
        ///             }
        /// 
        ///             Console.WriteLine("Table: row count = {0}", table.RowCount);
        /// 
        ///             //Inserting a valid value
        ///             table.InsertRow(column.Value(99));
        /// 
        ///             Console.WriteLine("Table: row count = {0}", table.RowCount);
        ///         }
        ///     }
        /// </code>
        /// </example>
        public int? Precision
        {
            get { return _precision ?? (_precision = DbValue.GetInteger(
                ReadProperty(Db.Tables.DecimalColumnList.Col.Precision))); }
            set
            {
                if (_precision.HasValue && _precision == value) return;

                Domain.DbAccess.ExecuteNonQuery(
                    cmd =>
                    {
                        cmd.CommandText = string.Format("{{call {0}(?,?)}}",
                            Domain.GetObjectFullName(Db.Procedures.SetDecimalPrecision));

                        cmd.Parameters.Add(Domain.DbAccess.CreateParameter("a_column_id",
                            OdbcType.Binary, GetRawId()));
                        cmd.Parameters.Add(Domain.DbAccess.CreateParameter("a_precision",
                            OdbcType.Decimal, DbValue.GetDbValue(value)));
                    });

                _precision = value;
            }
        }

        /// <summary>
        ///   Gets or sets the scale of the column.
        /// </summary>
        /// <remarks>
        /// <para>The scale of the column determines the number of digits after comma that can store a value in the column.
        /// The precision of the column must always be greater or equal to the scale of the column.
        /// The scale of the column is taken in account only if the precision of the column is also set.
        /// </para>
        /// <para>The current user has to have the permision to change the schema in order to set this property.</para>
        /// </remarks>
        /// <exception cref="System.Data.Common.DbException">Thrown when the current user has no permission to set this property.</exception>
        /// <exception cref="System.Data.Common.DbException">Thrown when trying to set the precision less than the scale.</exception>
        /// <example>
        ///   The following example shows how to set the scale of the column.
        /// <code>
        ///     using System;
        ///     using System.Data.Common;
        ///     using Centido.Core;
        /// 
        ///     class Sample
        ///     {
        ///         static void Main()
        ///         {
        ///             var domainConfig = new DomainConfiguration()
        ///             {
        ///                 DbConnectionString = @"Driver={SQL Server};Server=.\SQLEXPRESS;Database=MyDb;Uid=admin;Pwd=P@ssw0rd;",
        ///                 DbProviderName = "MSSQLSERVER"
        ///             };
        ///         
        ///             var domain = Domain.Build(domainConfig);
        ///         
        ///             domain.DeleteAll();
        ///             
        ///             //Creating a table
        ///             var table = domain.AddTable("my_table");
        ///             table.GrantAll();
        /// 
        ///             //Creating a decimal column
        ///             var column = table.AddColumn("my_column", ColumnType.Decimal)
        ///                 as DecimalColumn;
        ///             
        ///             //Setting the column precision
        ///             column.Precision = 2;
        ///             //Setting the column scale
        ///             column.Scale = 1;
        /// 
        ///             //Inserting a value with two digits after commma
        ///             var id = table.InsertRow(column.Value(0.25M));
        /// 
        ///             Console.WriteLine("Initial value = {0}, inserted value = {1}",
        ///                 0.25M, column.Read(id).GetDecimal());
        ///         }
        ///     }
        /// </code>
        /// </example>
        public int? Scale
        {
            get { return _scale ?? (_scale = DbValue.GetInteger(
                ReadProperty(Db.Tables.DecimalColumnList.Col.Scale))); }
            set
            {
                if (_scale.HasValue && _scale == value) return;

                Domain.DbAccess.ExecuteNonQuery(
                    cmd =>
                    {
                        cmd.CommandText = string.Format("{{call {0}(?,?)}}",
                            Domain.GetObjectFullName(Db.Procedures.SetDecimalScale));

                        cmd.Parameters.Add(Domain.DbAccess.CreateParameter("a_column_id",
                            OdbcType.Binary, GetRawId()));
                        cmd.Parameters.Add(Domain.DbAccess.CreateParameter("a_scale",
                            OdbcType.Decimal, DbValue.GetDbValue(value)));
                    });

                _scale = value;
            }
        }

        /// <summary>
        ///   Gets the column backend table name.
        /// </summary>
        protected override string BackEndTableName
        {
            get { return Domain.GetObjectFullName(Db.Tables.DecimalColumnList.Name); }
        }

        /// <summary>
        ///   Gets or sets the minimum value of the column.
        /// </summary>
        /// <remarks>The current user has to have the permission to change the schema in order to set this property.</remarks>
        /// <exception cref="System.Data.Common.DbException">Thrown when the current user has no permission to set this property.</exception>
        /// <exception cref="System.Data.Common.DbException">Thrown when trying to set this property to true and there are rows that contain invalid values.</exception>
        /// <example>
        ///   The following example shows how to set the minimum value of a decimal column.
        /// <code>
        ///     using System;
        ///     using System.Data.Common;
        ///     using Centido.Core;
        /// 
        ///     class Sample
        ///     {
        ///         static void Main()
        ///         {
        ///             var domainConfig = new DomainConfiguration()
        ///             {
        ///                 DbConnectionString = @"Driver={SQL Server};Server=.\SQLEXPRESS;Database=MyDb;Uid=admin;Pwd=P@ssw0rd;",
        ///                 DbProviderName = "MSSQLSERVER"
        ///             };
        ///         
        ///             var domain = Domain.Build(domainConfig);
        ///         
        ///             domain.DeleteAll();
        ///             
        ///             //Creating a table
        ///             var table = domain.AddTable("my_table");
        ///             table.GrantAll();
        /// 
        ///             //Creating a decimal column
        ///             var column = table.AddColumn("my_column", ColumnType.Decimal)
        ///                 as DecimalColumn;
        /// 
        ///             //Setting the column minimum
        ///             column.Minimum = 1;
        /// 
        ///             //Trying to insert an invalid value
        ///             try
        ///             {
        ///                 table.InsertRow(column.Value(0));
        ///             }
        ///             catch(DbException)
        ///             {
        ///                 Console.WriteLine("Inserting values less than the minimum is forbidden");
        ///             }
        /// 
        ///             Console.WriteLine("Table: row count = {0}", table.RowCount);
        /// 
        ///             //Inserting a valid value
        ///             table.InsertRow(column.Value(2));
        /// 
        ///             Console.WriteLine("Table: row count = {0}", table.RowCount);
        ///         }
        ///     }
        /// </code>
        /// </example>
        public decimal? Minimum
        {
            get { return _minimum ?? (_minimum = DbValue.GetDecimal(
                ReadProperty(Db.Tables.DecimalColumnList.Col.Minimum))); }
            set
            {
                if (_minimum.HasValue && _minimum == value) return;

                Domain.DbAccess.ExecuteNonQuery(
                    cmd =>
                    {
                        cmd.CommandText = string.Format("{{call {0}(?,?)}}",
                            Domain.GetObjectFullName(Db.Procedures.SetDecimalMinimum));

                        cmd.Parameters.Add(Domain.DbAccess.CreateParameter("a_column_id",
                            OdbcType.Binary, GetRawId()));
                        cmd.Parameters.Add(Domain.DbAccess.CreateParameter("a_minimum",
                            OdbcType.Decimal, DbValue.GetDbValue(value)));
                    });

                _minimum = value;
            }
        }

        /// <summary>
        ///   Gets or sets the maximum value of the column.
        /// </summary>
        /// <remarks>The current user has to have the permission to change the schema in order to set this property.</remarks>
        /// <exception cref="System.Data.Common.DbException">Thrown when the current user has no permission to set this property.</exception>
        /// <exception cref="System.Data.Common.DbException">Thrown when trying to set this property to true and there are rows that contain invalid values.</exception>
        /// <example>
        ///   The following example shows how to set the maximum value of a decimal column.
        /// <code>
        ///     using System;
        ///     using System.Data.Common;
        ///     using Centido.Core;
        /// 
        ///     class Sample
        ///     {
        ///         static void Main()
        ///         {
        ///             var domainConfig = new DomainConfiguration()
        ///             {
        ///                 DbConnectionString = @"Driver={SQL Server};Server=.\SQLEXPRESS;Database=MyDb;Uid=admin;Pwd=P@ssw0rd;",
        ///                 DbProviderName = "MSSQLSERVER"
        ///             };
        ///         
        ///             var domain = Domain.Build(domainConfig);
        ///         
        ///             domain.DeleteAll();
        ///             
        ///             //Creating a table
        ///             var table = domain.AddTable("my_table");
        ///             table.GrantAll();
        /// 
        ///             //Creating a decimal column
        ///             var column = table.AddColumn("my_column", ColumnType.Decimal)
        ///                 as DecimalColumn;
        /// 
        ///             //Setting the column maximum
        ///             column.Maximum = 1;
        /// 
        ///             //Trying to insert an invalid value
        ///             try
        ///             {
        ///                 table.InsertRow(column.Value(2));
        ///             }
        ///             catch(DbException)
        ///             {
        ///                 Console.WriteLine("Inserting values greater than the maximum is forbidden");
        ///             }
        /// 
        ///             Console.WriteLine("Table: row count = {0}", table.RowCount);
        /// 
        ///             //Inserting a valid value
        ///             table.InsertRow(column.Value(0));
        /// 
        ///             Console.WriteLine("Table: row count = {0}", table.RowCount);
        ///         }
        ///     }
        /// </code>
        /// </example>
        public decimal? Maximum
        {
            get { return _maximum ?? (_maximum = DbValue.GetDecimal(
                ReadProperty(Db.Tables.DecimalColumnList.Col.Maximum))); }
            set
            {
                if (_maximum.HasValue && _maximum == value) return;

                Domain.DbAccess.ExecuteNonQuery(
                    cmd =>
                    {
                        cmd.CommandText = string.Format("{{call {0}(?,?)}}",
                            Domain.GetObjectFullName(Db.Procedures.SetDecimalMaximum));

                        cmd.Parameters.Add(Domain.DbAccess.CreateParameter("a_column_id",
                            OdbcType.Binary, GetRawId()));
                        cmd.Parameters.Add(Domain.DbAccess.CreateParameter("a_maximum",
                            OdbcType.Decimal, DbValue.GetDbValue(value)));
                    });

                _maximum = value;
            }
        }


        /// <summary>
        ///   Gets or sets the default value of the column.
        /// </summary>
        /// <remarks>The current user has to have the permission to change the schema in order to set this property.</remarks>
        /// <exception cref="System.Data.Common.DbException">Thrown when the current user has no permission to set this property.</exception>
        /// <example>
        ///   The following example shows how to set the default value of a decimal column.
        /// <code>
        ///     using System;
        ///     using Centido.Core;
        /// 
        ///     class Sample
        ///     {
        ///         static void Main()
        ///         {
        ///             var domainConfig = new DomainConfiguration()
        ///             {
        ///                 DbConnectionString = @"Driver={SQL Server};Server=.\SQLEXPRESS;Database=MyDb;Uid=admin;Pwd=P@ssw0rd;",
        ///                 DbProviderName = "MSSQLSERVER"
        ///             };
        ///         
        ///             var domain = Domain.Build(domainConfig);
        ///         
        ///             domain.DeleteAll();
        ///             
        ///             //Creating a table
        ///             var table = domain.AddTable("my_table");
        ///             table.GrantAll();
        /// 
        ///             //Creating a decimal column
        ///             var column = table.AddColumn("my_column", ColumnType.Decimal)
        ///                 as DecimalColumn;
        ///             column.Precision = 10;
        ///             column.Scale = 3;
        /// 
        ///             //Setting the column default value
        ///             column.DefaultValue = 1.5M;
        /// 
        ///             //Inserting a row
        ///             var id = table.InsertRow();
        /// 
        ///             Console.WriteLine("Inserted value = {0}", column.Read(id).GetDecimal());
        ///         }
        ///     }
        /// </code>
        /// </example>
        public decimal? DefaultValue
        {
            get { return _defaultValue ?? (_defaultValue = DbValue.GetDecimal(
                ReadProperty(Db.Tables.DecimalColumnList.Col.DefaultValue))); }
            set
            {
                if (_defaultValue.HasValue && _defaultValue == value) return;

                Domain.DbAccess.ExecuteNonQuery(
                    cmd =>
                    {
                        cmd.CommandText = string.Format("{{call {0}(?,?)}}",
                            Domain.GetObjectFullName(Db.Procedures.SetDecimalDefaultValue));

                        cmd.Parameters.Add(Domain.DbAccess.CreateParameter("a_column_id",
                            OdbcType.Binary, GetRawId()));
                        cmd.Parameters.Add(Domain.DbAccess.CreateParameter("a_default_value",
                            OdbcType.Decimal, DbValue.GetDbValue(value)));
                    });

                _defaultValue = value;
            }
        }


        /// <summary>
        ///   Refreshes the properties of the column.
        /// </summary>
        public override void Refresh()
        {
            base.Refresh();

            _defaultValue = null;
            _maximum = null;
            _minimum = null;
            _scale = null;
            _precision = null;
        }


        /// <summary>
        ///   Adds a new decimal column with the specified name to the specified table.
        /// </summary>
        /// <remarks>The current user has to have the permission to change the schema in order to execute this function.</remarks>
        /// <param name="table">The table.</param>
        /// <param name="columnName">The column name.</param>
        /// <exception cref="ArgumentNullException">Thrown when the value of <paramref name="table"/> is null.</exception>
        /// <exception cref="System.ArgumentNullException">Thrown when the value of <paramref name="columnName"/> is null.</exception>
        /// <exception cref="System.ArgumentException">Thrown when the value of <paramref name="columnName"/> is an empty string.</exception>
        /// <exception cref="System.ArgumentException">Thrown when the value of <paramref name="columnName"/> is not a valid identifier.</exception>
        /// <exception cref="System.Data.Common.DbException">Thrown the current user has no permission to execute this function.</exception>
        /// <exception cref="System.Data.Common.DbException">Thrown when there is a column with the same name in the same user-defined table.</exception>
        /// <returns>Returns the newly created decimal table column.</returns>
        /// <example>
        ///   The following example shows how to add a decimal column.
        /// <code>
        ///     using System;
        ///     using Centido.Core;
        /// 
        ///     class Sample
        ///     {
        ///         static void Main()
        ///         {
        ///             var domainConfig = new DomainConfiguration()
        ///             {
        ///                 DbConnectionString = @"Driver={SQL Server};Server=.\SQLEXPRESS;Database=MyDb;Uid=admin;Pwd=P@ssw0rd;",
        ///                 DbProviderName = "MSSQLSERVER"
        ///             };
        ///         
        ///             var domain = Domain.Build(domainConfig);
        ///             
        ///             domain.DeleteAll();
        ///             
        ///             //Creating a new table
        ///             var table = domain.AddTable("my_table");
        /// 
        ///             //Creating a decimal table column
        ///             var column = DecimalColumn.Add(table, "my_column");
        ///         }
        ///     }
        /// </code>
        /// </example>
        public static DecimalColumn Add(Table table, string columnName)
        {
            if (table == null)
                throw new ArgumentNullException("table");

            return table.AddColumn(columnName, ColumnType.Decimal) as DecimalColumn;
        }


        /// <summary>
        ///   Adds a new decimal table column with the specified id and name to the specified table.
        /// </summary>
        /// <remarks>The current user has to have the permission to change the schema in order to execute this function.</remarks>
        /// <param name="table">The table.</param>
        /// <param name="columnId">The column id.</param>
        /// <param name="columnName">The column name.</param>
        /// <exception cref="ArgumentNullException">Thrown when the value of <paramref name="table"/> is null.</exception>
        /// <exception cref="System.ArgumentNullException">Thrown when the value of <paramref name="columnName"/> is null.</exception>
        /// <exception cref="System.ArgumentException">Thrown when the value of <paramref name="columnName"/> is an empty string.</exception>
        /// <exception cref="System.ArgumentException">Thrown when the value of <paramref name="columnName"/> is not a valid identifier.</exception>
        /// <exception cref="System.Data.Common.DbException">Thrown the current user has no permission to execute this function.</exception>
        /// <exception cref="System.Data.Common.DbException">Thrown when there is a column with the same id or the same name in the same user-defined table.</exception>
        /// <returns>Returns the newly created decimal table column.</returns>
        /// <example>
        ///   The following example shows how to create a decimal table column.
        /// <code>
        ///     using System;
        ///     using Centido.Core;
        /// 
        ///     class Sample
        ///     {
        ///         static void Main()
        ///         {
        ///             var domainConfig = new DomainConfiguration()
        ///             {
        ///                 DbConnectionString = @"Driver={SQL Server};Server=.\SQLEXPRESS;Database=MyDb;Uid=admin;Pwd=P@ssw0rd;",
        ///                 DbProviderName = "MSSQLSERVER"
        ///             };
        ///         
        ///             var domain = Domain.Build(domainConfig);
        ///             
        ///             domain.DeleteAll();
        ///             
        ///             //Creating a new table
        ///             var table = domain.AddTable("my_table");
        /// 
        ///             //Creating a decimal table column
        ///             var column = DecimalColumn.Add(table, Guid.NewGuid(), "my_column");
        ///         }
        ///     }
        /// </code>
        /// </example>
        public static DecimalColumn Add(Table table, Guid columnId, string columnName)
        {
            if (table == null)
                throw new ArgumentNullException("table");

            return table.AddColumn(columnId, columnName, ColumnType.Decimal) as DecimalColumn;
        }


        /// <summary>
        ///   Ensures that the table contains a decimal column with the specified name.
        /// </summary>
        /// <remarks>The current user has to have the permission to change the schema in order to execute this function.</remarks>
        /// <param name="table">The table.</param>
        /// <param name="columnName">The column name.</param>
        /// <exception cref="ArgumentNullException">Thrown when the value of <paramref name="table"/> is null.</exception>
        /// <exception cref="System.Data.Common.DbException">Thrown when trying to add a column and
        /// the current user has no permission to change the schema.</exception>
        /// <exception cref="ArgumentNullException">Thrown when the value of <paramref name="columnName"/> is null.</exception>
        /// <exception cref="ArgumentException">Thrown when the value of <paramref name="columnName"/> is an empty string.</exception>
        /// <exception cref="ArgumentException">Thrown when the value of <paramref name="columnName"/> is not a valid identifier.</exception>
        /// <exception cref="DbOperationException">Thrown when the table already has a column with the 
        /// specified name but different column type.</exception>
        /// <returns>Returns a decimal table column with the specified name.</returns>
        /// <example>
        ///   The following example shows how to ensure that a table has a decimal column with the specified name.
        /// <code>
        ///     using System;
        ///     using Centido.Core;
        /// 
        ///     class Sample
        ///     {
        ///         static void Main()
        ///         {
        ///             var domainConfig = new DomainConfiguration()
        ///             {
        ///                 DbConnectionString = @"Driver={SQL Server};Server=.\SQLEXPRESS;Database=MyDb;Uid=admin;Pwd=P@ssw0rd;",
        ///                 DbProviderName = "MSSQLSERVER"
        ///             };
        ///         
        ///             var domain = Domain.Build(domainConfig);
        ///             
        ///             domain.DeleteAll();
        ///             
        ///             //Creating a new table
        ///             var table = domain.AddTable("my_table");
        /// 
        ///             //Creating a decimal table column
        ///             table.AddColumn("first_column", ColumnType.Decimal);
        /// 
        ///             //Ensuring the first column
        ///             DecimalColumn.Ensure(table, "first_column");
        ///             //Ensuring the second column
        ///             DecimalColumn.Ensure(table, "second_column");
        /// 
        ///             foreach(var column in table.Columns)
        ///                 Console.WriteLine("Table column: name = {0}, type = {1}", column.Name, column.ColumnType);
        ///         }
        ///     }
        /// </code>
        /// </example>
        public static DecimalColumn Ensure(Table table, string columnName)
        {
            if (table == null)
                throw new ArgumentNullException("table");

            return table.EnsureColumn(columnName, ColumnType.Decimal) as DecimalColumn;
        }


        /// <summary>
        ///   Ensures that the table contains a decimal column with the specified id and the specified name.
        /// </summary>
        /// <param name="table">The table.</param>
        /// <param name="columnId">The column id.</param>
        /// <param name="columnName">The column name.</param>
        /// <exception cref="ArgumentNullException">Thrown when the value of <paramref name="table"/> is null.</exception>
        /// <exception cref="System.Data.Common.DbException">Thrown when trying to add a column and
        /// the current user has no permission to change the schema.</exception>
        /// <exception cref="ArgumentNullException">Thrown when the value of <paramref name="columnName"/> is null.</exception>
        /// <exception cref="ArgumentException">Thrown when the value of <paramref name="columnName"/> is an empty string.</exception>
        /// <exception cref="ArgumentException">Thrown when the value of <paramref name="columnName"/> is not a valid identifier.</exception>
        /// <exception cref="DbOperationException">Thrown when the table already has a column with the
        /// specified id but different column name.</exception>
        /// <exception cref="DbOperationException">Thrown when the table already has a column with the 
        /// specified id but different column type.</exception>
        /// <exception cref="DbOperationException">Thrown when the table already has a column with the specified name
        /// but different column id.</exception>
        /// <returns>Returns the table column with the specified id, the specified name and the specified type.</returns>
        /// <returns>Returns the table column with the specified id, the specified name and the specified type.</returns>
        /// <example>
        ///   The following example shows how to ensure a decimal column with the specified id and
        /// the specified name.
        /// <code>
        ///     using System;
        ///     using Centido.Core;
        /// 
        ///     class Sample
        ///     {
        ///         static void Main()
        ///         {
        ///             var domainConfig = new DomainConfiguration()
        ///             {
        ///                 DbConnectionString = @"Driver={SQL Server};Server=.\SQLEXPRESS;Database=MyDb;Uid=admin;Pwd=P@ssw0rd;",
        ///                 DbProviderName = "MSSQLSERVER"
        ///             };
        ///         
        ///             var domain = Domain.Build(domainConfig);
        ///             
        ///             domain.DeleteAll();
        ///             
        ///             //Creating a new table
        ///             var table = domain.AddTable("my_table");
        /// 
        ///             //Creating a table column
        ///             var columnId = Guid.NewGuid();
        ///             table.AddColumn(columnId, "first_column", ColumnType.Decimal);
        /// 
        ///             //Ensuring the first column
        ///             DecimalColumn.Ensure(table, columnId, "first_column");
        ///             //Ensuring the second column
        ///             DecimalColumn.Ensure(table, Guid.NewGuid(), "second_column");
        /// 
        ///             foreach(var column in table.Columns)
        ///                 Console.WriteLine("Table column: name = {0}, type = {1}", column.Name, column.ColumnType);
        ///         }
        ///     }
        /// </code>
        /// </example>
        public static DecimalColumn Ensure(Table table, Guid columnId, string columnName)
        {
            if (table == null)
                throw new ArgumentNullException("table");

            return table.EnsureColumn(columnId, columnName, ColumnType.Decimal) as DecimalColumn;
        }
    }
}