﻿using System;
using System.Data.Common;
using System.Linq;

namespace Centido.Management.ConsoleManager
{
    /// <summary>
    ///   A console command that deletes the specified business unit from the current domain
    /// or removes the unit from the specified parent business unit.
    /// </summary>
    internal class DeleteUnitCommand : Command
    {
        /// <summary>
        ///   The unit name parameter.
        /// </summary>
        public const string UnitNameParameter = "unit";

        /// <summary>
        ///   The unit name parameter (short name).
        /// </summary>
        public const string UnitNameShortParameter = "u";

        /// <summary>
        ///   The parent unit name parameter.
        /// </summary>
        public const string ParentUnitNameParameter = "parentunit";

        /// <summary>
        ///   The parent unit name parameter (short name).
        /// </summary>
        public const string ParentUnitNameShortParameter = "pu";


        /// <summary>
        ///   Gets the console command name.
        /// </summary>
        public override string Name
        {
            get { return "deleteunit"; }
        }


        /// <summary>
        ///   Gets the console command description.
        /// </summary>
        public override string Description
        {
            get { return Messages.DeleteUnitCommandDescription; }
        }


        /// <summary>
        ///   Gets the console command full description.
        /// </summary>
        public override string FullDescription
        {
            get { return Messages.DeleteUnitCommandFullDescription; }
        }


        /// <summary>
        ///   Executes the console command.
        /// </summary>
        /// <param name="context">The console command execution context.</param>
        /// <exception cref="ArgumentNullException">Thrown when the value of <paramref name="context"/> is null.</exception>
        public override void Execute(CommandExecutionContext context)
        {
            base.Execute(context);

            // If the unit name was not specified
            if (!context.ParameterValueIsNotEmpty(UnitNameParameter, UnitNameShortParameter))
            {
                Console.WriteLine(Messages.DeleteUnitCommandNotAllParametersSpecifiedError);
                return;
            }

            var domain = context.CurrentDomain;

            if (domain == null) return;

            // If the current user has no permission to perform this action
            if (!domain.CurrentUser.CanChangeSecurity)
            {
                Console.WriteLine(Messages.NeedCanChangeSecurityPermissionMessage);
                return;
            }

            // The unit name
            var unitName = context.GetParameterValue(UnitNameParameter, UnitNameShortParameter);

            try
            {
                // Trying to find the unit with the specified name
                var unit = domain.GetUnit(unitName);

                // If the unit with the specified name was not found
                if (unit == null)
                {
                    Console.WriteLine(Messages.CannotFindUnitError, unitName);
                    return;
                }

                // If the parent unit was specified
                if (context.ParameterValueIsNotEmpty(ParentUnitNameParameter, ParentUnitNameShortParameter))
                {
                    // ... then deleting the unit from the parent unit

                    // The parent unit name
                    var parentUnitName = context.GetParameterValue(ParentUnitNameParameter, ParentUnitNameShortParameter);

                    // Trying to find the parent unit
                    var parentUnit = domain.GetUnit(parentUnitName);

                    // If there is no unit with the specified name
                    if (parentUnit == null)
                    {
                        Console.WriteLine(Messages.CannotFindUnitError, parentUnitName);
                        return;
                    }

                    if (parentUnit.ChildUnits.Any(u => u.Id == unit.Id))
                        parentUnit.RemoveUnit(unit);
                }
                // If the parent unit was not specified
                else
                {
                    // ... then deleting the unit
                    unit.Delete();
                }
            }
            catch (DbException e)
            {
                Console.WriteLine(e.Message);
                Console.WriteLine();
                return;
            }

            // Displaying the success message
            Console.WriteLine(Messages.CommandCompletedMessage);
        }
    }
}
