﻿// -----------------------------------------------------------------------
// <copyright file="DecimalColumnDto.cs" company="JSC Centido">
//  Copyright (c) JSC Centido. All rights reserved.
// </copyright>
// -----------------------------------------------------------------------

using System;
using System.Globalization;
using System.Xml;
using System.Xml.Schema;

namespace Centido.Core.Dto
{
    /// <summary>
    ///   A data transfer object for a decimal table column.
    /// </summary>
    public class DecimalColumnDto : ColumnDto
    {
        /// <summary>
        ///   The name of the xml tag that is used to store this object.
        /// </summary>
        public const string XmlTag = "decimalColumn";

        /// <summary>
        ///   The current xml schema version.
        /// </summary>
        public const string CurrentXmlSchemaVersion = "1";

        /// <summary>
        ///   The name of the xml attribute that is used to store the column default value.
        /// </summary>
        public const string DefaultValueAttribute = "defaultValue";

        /// <summary>
        ///   The name of the xml attribute that is used to store the column maximum value.
        /// </summary>
        public const string MaximumAttribute = "maxValue";

        /// <summary>
        ///   The name of the xml attribute that is used to store the column minimum value.
        /// </summary>
        public const string MinimumAttribute = "minValue";

        /// <summary>
        ///   The name of the xml attribute that is used to store the column precision.
        /// </summary>
        public const string PrecisionAttribute = "precision";

        /// <summary>
        ///   The name of the xml attribute that is used to store the column scale.
        /// </summary>
        public const string ScaleAttribute = "scale";


        /// <summary>
        ///   Gets the column type.
        /// </summary>
        public override ColumnType ColumnType
        {
            get { return ColumnType.Decimal; }
        }


        /// <summary>
        ///   Gets or sets the column default value.
        /// </summary>
        public decimal? DefaultValue { get; set; }


        /// <summary>
        ///   Gets or sets the column maximum value.
        /// </summary>
        public decimal? Maximum { get; set; }


        /// <summary>
        ///   Gets or sets the column minimum value.
        /// </summary>
        public decimal? Minimum { get; set; }


        /// <summary>
        ///   Gets or sets the column precision.
        /// </summary>
        public int? Precision { get; set; }


        /// <summary>
        ///   Gets or sets the column scale.
        /// </summary>
        public int? Scale { get; set; }


        /// <summary>
        ///   This method is reserved and should not be used. When implementing the IXmlSerializable interface, you should return null (Nothing in Visual Basic) from this method, and instead, if specifying a custom schema is required, apply the <see cref="T:System.Xml.Serialization.XmlSchemaProviderAttribute"/> to the class.
        /// </summary>
        /// <returns>
        ///   An <see cref="T:System.Xml.Schema.XmlSchema"/> that describes the XML representation of the object that is produced by the <see cref="M:System.Xml.Serialization.IXmlSerializable.WriteXml(System.Xml.XmlWriter)"/> method and consumed by the <see cref="M:System.Xml.Serialization.IXmlSerializable.ReadXml(System.Xml.XmlReader)"/> method.
        /// </returns>
        public override XmlSchema GetSchema()
        {
            return null;
        }


        /// <summary>
        ///   Generates an object from its XML representation.
        /// </summary>
        /// <param name="reader">The <see cref="T:System.Xml.XmlReader"/> stream from which the object is deserialized. 
        /// </param>
        /// <exception cref="ArgumentNullException">Thrown when the value of <paramref name="reader"/> is null.</exception>
        public override void ReadXml(XmlReader reader)
        {
            if (reader == null)
                throw new ArgumentNullException("reader");

            reader.MoveToContent();

            // Reading the xml schema version
            var xmlSchemaVersion = ReadStringAttribute(reader, XmlSchemaVersionTag, isRequired: false);

            // If the xml schema version is not defined
            if (string.IsNullOrEmpty(xmlSchemaVersion)
                // or if it is the current xml schema version
                || xmlSchemaVersion == CurrentXmlSchemaVersion)
            {
                // Reading the column name
                Name = ReadStringAttribute(reader, NameAttribute);

                // Reading the display name
                DisplayName = ReadStringAttribute(reader, DisplayNameAttribute, isRequired: false);

                // Reading the description
                Description = ReadStringAttribute(reader, DescriptionAttribute, isRequired: false);

                // Reading the order index
                OrderIndex = ReadIntAttribute(reader, OrderIndexAttribute, isRequired: false);

                // Reading the column nullability
                IsNotNull = ReadBoolAttribute(reader, IsNotNullAttribute, isRequired: false);

                // Reading the default value
                DefaultValue = ReadDecimalAttribute(reader, DefaultValueAttribute, isRequired: false);

                // Reading the maximum value
                Maximum = ReadDecimalAttribute(reader, MaximumAttribute, isRequired: false);

                // Reading the minimum value
                Minimum = ReadDecimalAttribute(reader, MinimumAttribute, isRequired: false);

                // Reading the precision
                Precision = ReadIntAttribute(reader, PrecisionAttribute, isRequired: false);

                // Reading the scale
                Scale = ReadIntAttribute(reader, ScaleAttribute, isRequired: false);

                var isEmptyElement = reader.IsEmptyElement;

                reader.ReadStartElement();

                if (!isEmptyElement)
                    reader.ReadEndElement();
            }
            else
                throw new ArgumentException(string.Format(
                    "The specified xml schema version is not supported (node = {0}, version = {1}).",
                    reader.Name,
                    xmlSchemaVersion));
        }


        /// <summary>
        ///   Converts an object into its XML representation.
        /// </summary>
        /// <param name="writer">The <see cref="T:System.Xml.XmlWriter"/> stream to which the object is serialized. 
        /// </param>
        /// <exception cref="ArgumentNullException">Thrown when the value of <paramref name="writer"/> is null.</exception>
        public override void WriteXml(XmlWriter writer)
        {
            if (writer == null)
                throw new ArgumentNullException("writer");

            writer.WriteStartElement(XmlTag);

            // The xml schema version
            writer.WriteAttributeString(XmlSchemaVersionTag, CurrentXmlSchemaVersion);

            // The column name
            if (!string.IsNullOrEmpty(Name))
                writer.WriteAttributeString(NameAttribute, Name);

            // The column display name
            if (!string.IsNullOrEmpty(DisplayName))
                writer.WriteAttributeString(DisplayNameAttribute, DisplayName);

            // The column description
            if (!string.IsNullOrEmpty(Description))
                writer.WriteAttributeString(DescriptionAttribute, Description);

            // The column order index
            if (OrderIndex.HasValue)
                writer.WriteAttributeString(OrderIndexAttribute, 
                    OrderIndex.Value.ToString(CultureInfo.InvariantCulture));

            // The column nullability
            if (IsNotNull.HasValue)
                writer.WriteAttributeString(IsNotNullAttribute,
                    IsNotNull.Value.ToString(CultureInfo.InvariantCulture));

            // The column default value
            if (DefaultValue.HasValue)
                writer.WriteAttributeString(DefaultValueAttribute,
                    DefaultValue.Value.ToString(CultureInfo.InvariantCulture));

            // The column maximum value
            if (Maximum.HasValue)
                writer.WriteAttributeString(MaximumAttribute,
                    Maximum.Value.ToString(CultureInfo.InvariantCulture));

            // The column minimum value
            if (Minimum.HasValue)
                writer.WriteAttributeString(MinimumAttribute,
                    Minimum.Value.ToString(CultureInfo.InvariantCulture));

            // The column precision
            if (Precision.HasValue)
                writer.WriteAttributeString(PrecisionAttribute,
                    Precision.Value.ToString(CultureInfo.InstalledUICulture));

            // The column scale
            if (Scale.HasValue)
                writer.WriteAttributeString(ScaleAttribute,
                    Scale.Value.ToString(CultureInfo.InstalledUICulture));

            writer.WriteEndElement();
        }


        /// <summary>
        ///   Copies the properties of the data transfer object to the specified decimal column.
        /// </summary>
        /// <param name="column">The decimal column.</param>
        /// <exception cref="ArgumentNullException">Thrown when the value of <paramref name="column"/> is null.</exception>
        /// <exception cref="ArgumentException">Thrown when the column has the type other than decimal.</exception>
        public override void CopyTo(Column column)
        {
            if (column == null)
                throw new ArgumentNullException("column");
            if (column.ColumnType != ColumnType.Decimal)
                throw new ArgumentException("The column type should be decimal.");

            var decimalColumn = (DecimalColumn)column;

            // The column name
            if (!string.IsNullOrEmpty(Name) && Name != column.Name)
                column.Name = Name;

            // The column display name
            if (DisplayName != column.DisplayName)
                column.DisplayName = DisplayName;

            // The column description
            if (Description != column.Description)
                column.Description = Description;

            // The column order index
            if (OrderIndex.HasValue && OrderIndex != column.OrderIndex)
                column.OrderIndex = OrderIndex.Value;

            // The column nullability
            if (IsNotNull.HasValue && IsNotNull != column.IsNotNull)
                column.IsNotNull = IsNotNull.Value;

            // The column default value
            if (DefaultValue != decimalColumn.DefaultValue)
                decimalColumn.DefaultValue = DefaultValue;

            // The column maximum value
            if (Maximum != decimalColumn.Maximum)
                decimalColumn.Maximum = Maximum;

            // The column minimum value
            if (Minimum != decimalColumn.Minimum)
                decimalColumn.Minimum = Minimum;

            // The column precision
            if (Precision != decimalColumn.Precision)
                decimalColumn.Precision = Precision;

            // The column scale
            if (Scale != decimalColumn.Scale)
                decimalColumn.Scale = Scale;
        }


        /// <summary>
        ///   Creates a data transfer object for the specified decimal column.
        /// </summary>
        /// <param name="decimalColumn">The specified decimal column.</param>
        /// <exception cref="ArgumentNullException">Thrown when the value of <paramref name="decimalColumn"/> is null.</exception>
        /// <returns>Returns the data transfer object for the specified decimal column.</returns>
        public static DecimalColumnDto CreateFrom(DecimalColumn decimalColumn)
        {
            if (decimalColumn == null)
                throw new ArgumentNullException("decimalColumn");

            return new DecimalColumnDto
            {
                Name = decimalColumn.Name,
                DisplayName = decimalColumn.DisplayName,
                Description = decimalColumn.Description,
                OrderIndex = decimalColumn.OrderIndex,
                IsNotNull = decimalColumn.IsNotNull,
                DefaultValue = decimalColumn.DefaultValue,
                Maximum = decimalColumn.Maximum,
                Minimum = decimalColumn.Minimum,
                Precision = decimalColumn.Precision,
                Scale = decimalColumn.Scale
            };
        }
    }
}
