﻿// -----------------------------------------------------------------------
// <copyright file="EnumColumnDto.cs" company="JSC Centido">
//  Copyright (c) JSC Centido. All rights reserved.
// </copyright>
// -----------------------------------------------------------------------

using System;
using System.Collections.Generic;
using System.Globalization;
using System.Xml;
using System.Xml.Schema;

namespace Centido.Core.Dto
{
    /// <summary>
    ///   A data transfer object for an enumeration table column.
    /// </summary>
    public class EnumColumnDto : ColumnDto
    {
        /// <summary>
        ///   The name of the xml tag that is used to store this object.
        /// </summary>
        public const string XmlTag = "enumColumn";

        /// <summary>
        ///   The current xml schema version.
        /// </summary>
        public const string CurrentXmlSchemaVersion = "1";

        /// <summary>
        ///   The name of the xml attribute that is used to store a column default value.
        /// </summary>
        public const string DefaultValueAttribute = "defaultValue";

        /// <summary>
        ///   The name of the xml attribute that is used to store the availability of the enumeration value checking.
        /// </summary>
        public const string HasValueCheckingAttribute = "hasValueChecking";

        /// <summary>
        ///   The name of the xml attribute that is used to store the enumeration member list.
        /// </summary>
        public const string MemberListAttribute = "members";


        /// <summary>
        ///   The list enumeration column members.
        /// </summary>
        private readonly List<EnumMemberDto> _enumMembers = new List<EnumMemberDto>();


        /// <summary>
        ///   Gets the column type.
        /// </summary>
        public override ColumnType ColumnType
        {
            get { return ColumnType.Enumeration; }
        }


        /// <summary>
        ///   Gets or sets the default value of the column.
        /// </summary>
        public int? DefaultValue { get; set; }

        /// <summary>
        ///   Gets or sets the availability of the enumeration value checking.
        /// </summary>
        public bool? HasValueChecking { get; set; }


        /// <summary>
        ///   Gets the enumeration column members.
        /// </summary>
        public ICollection<EnumMemberDto> Members
        {
            get { return _enumMembers; }
        }


        /// <summary>
        ///   Adds the specified enumeration column member.
        /// </summary>
        /// <param name="memberDto">The enumeration column member to be added.</param>
        /// <exception cref="ArgumentNullException">Thrown when the value of <paramref name="memberDto"/> is null.</exception>
        public void AddMember(EnumMemberDto memberDto)
        {
            if (memberDto == null)
                throw new ArgumentNullException("memberDto");

            if (!_enumMembers.Contains(memberDto))
                _enumMembers.Add(memberDto);
        }


        /// <summary>
        ///   Removes the specified enumeration member.
        /// </summary>
        /// <param name="memberDto">The enumeration column member to be removed.</param>
        /// <exception cref="ArgumentNullException">Thrown when the value of <paramref name="memberDto"/> is null.</exception>
        public void DeleteMember(EnumMemberDto memberDto)
        {
            if (memberDto == null)
                throw new ArgumentNullException("memberDto");

            if (_enumMembers.Contains(memberDto))
                _enumMembers.Remove(memberDto);
        }


        /// <summary>
        ///   This method is reserved and should not be used. When implementing the IXmlSerializable interface, you should return null (Nothing in Visual Basic) from this method, and instead, if specifying a custom schema is required, apply the <see cref="T:System.Xml.Serialization.XmlSchemaProviderAttribute"/> to the class.
        /// </summary>
        /// <returns>
        ///   An <see cref="T:System.Xml.Schema.XmlSchema"/> that describes the XML representation of the object that is produced by the <see cref="M:System.Xml.Serialization.IXmlSerializable.WriteXml(System.Xml.XmlWriter)"/> method and consumed by the <see cref="M:System.Xml.Serialization.IXmlSerializable.ReadXml(System.Xml.XmlReader)"/> method.
        /// </returns>
        public override XmlSchema GetSchema()
        {
            return null;
        }


        /// <summary>
        ///   Generates an object from its XML representation.
        /// </summary>
        /// <param name="reader">The <see cref="T:System.Xml.XmlReader"/> stream from which the object is deserialized. 
        /// </param>
        /// <exception cref="ArgumentNullException">Thrown when the value of <paramref name="reader"/> is null.</exception>
        public override void ReadXml(XmlReader reader)
        {
            if (reader == null)
                throw new ArgumentNullException("reader");

            reader.MoveToContent();

            // Reading the xml schema version
            var xmlSchemaVersion = ReadStringAttribute(reader, XmlSchemaVersionTag, isRequired: false);

            // If the xml schema version is not defined
            if (string.IsNullOrEmpty(xmlSchemaVersion)
                // or if it is the current xml schema version
                || xmlSchemaVersion == CurrentXmlSchemaVersion)
            {
                // Reading the column name
                Name = ReadStringAttribute(reader, NameAttribute);

                // Reading the display name
                DisplayName = ReadStringAttribute(reader, DisplayNameAttribute, isRequired: false);

                // Reading the description
                Description = ReadStringAttribute(reader, DescriptionAttribute, isRequired: false);

                // Reading the order index
                OrderIndex = ReadIntAttribute(reader, OrderIndexAttribute, isRequired: false);

                // Reading the column nullability
                IsNotNull = ReadBoolAttribute(reader, IsNotNullAttribute, isRequired: false);

                // Reading the default value
                DefaultValue = ReadIntAttribute(reader, DefaultValueAttribute, isRequired: false);

                // Reading the value checking
                HasValueChecking = ReadBoolAttribute(reader, HasValueCheckingAttribute, isRequired: false);

                // Clearing the list of members
                _enumMembers.Clear();

                var isEmptyElement = reader.IsEmptyElement;

                reader.ReadStartElement();

                if (!isEmptyElement)
                {
                    // The member list is specified
                    if (reader.LocalName == MemberListAttribute)
                    {
                        isEmptyElement = reader.IsEmptyElement;

                        reader.ReadStartElement();

                        if (!isEmptyElement)
                        {
                            // Reading the enumeration column members
                            while (reader.LocalName == EnumMemberDto.XmlTag)
                            {
                                var enumMember = new EnumMemberDto();

                                enumMember.ReadXml(reader);

                                _enumMembers.Add(enumMember);
                            }

                            reader.ReadEndElement();
                        }
                    }

                    reader.ReadEndElement();
                }
            }
            else
                throw new ArgumentException(string.Format(
                    "The specified xml schema version is not supported (node = {0}, version = {1}).",
                    reader.Name,
                    xmlSchemaVersion));
        }


        /// <summary>
        ///   Converts an object into its XML representation.
        /// </summary>
        /// <param name="writer">The <see cref="T:System.Xml.XmlWriter"/> stream to which the object is serialized. 
        /// </param>
        /// <exception cref="ArgumentNullException">Thrown when the value of <paramref name="writer"/> is null.</exception>
        public override void WriteXml(XmlWriter writer)
        {
            if (writer == null)
                throw new ArgumentNullException("writer");

            writer.WriteStartElement(XmlTag);

            // The xml schema version
            writer.WriteAttributeString(XmlSchemaVersionTag, CurrentXmlSchemaVersion);

            // The column name
            if (!string.IsNullOrEmpty(Name))
                writer.WriteAttributeString(NameAttribute, Name);

            // The column display name
            if (!string.IsNullOrEmpty(DisplayName))
                writer.WriteAttributeString(DisplayNameAttribute, DisplayName);

            // The column description
            if (!string.IsNullOrEmpty(Description))
                writer.WriteAttributeString(DescriptionAttribute, Description);

            // The column order index
            if (OrderIndex.HasValue)
                writer.WriteAttributeString(OrderIndexAttribute, 
                    OrderIndex.Value.ToString(CultureInfo.InvariantCulture));

            // The column nullability
            if (IsNotNull.HasValue)
                writer.WriteAttributeString(IsNotNullAttribute,
                    IsNotNull.Value.ToString(CultureInfo.InvariantCulture));

            // The column default value
            if (DefaultValue.HasValue)
                writer.WriteAttributeString(DefaultValueAttribute,
                    DefaultValue.Value.ToString(CultureInfo.InvariantCulture));

            // The column value checking
            if (HasValueChecking.HasValue)
                writer.WriteAttributeString(HasValueCheckingAttribute,
                    HasValueChecking.Value.ToString(CultureInfo.InvariantCulture));

            // If there are enumeration column members
            if (_enumMembers.Count > 0)
            {
                writer.WriteStartElement(MemberListAttribute);

                // Writing the enumeration column members
                foreach (var memberDto in _enumMembers)
                    memberDto.WriteXml(writer);

                writer.WriteEndElement();
            }

            writer.WriteEndElement();
        }


        /// <summary>
        ///   Copies the properties of the data transfer object to the specified enumeration column.
        /// </summary>
        /// <param name="column">The enumeration column.</param>
        /// <exception cref="ArgumentNullException">Thrown when the value of <paramref name="column"/> is null.</exception>
        /// <exception cref="ArgumentException">Thrown when the column has the type other than enumeration.</exception>
        public override void CopyTo(Column column)
        {
            if (column == null)
                throw new ArgumentNullException("column");
            if (column.ColumnType != ColumnType.Enumeration)
                throw new ArgumentException("The column type should be enumeration.");

            var enumColumn = (EnumColumn)column;

            // The column name
            if (!string.IsNullOrEmpty(Name) && Name != column.Name)
                column.Name = Name;

            // The column display name
            if (DisplayName != column.DisplayName)
                column.DisplayName = DisplayName;

            // The column description
            if (Description != column.Description)
                column.Description = Description;

            // The column order index
            if (OrderIndex.HasValue && OrderIndex != column.OrderIndex)
                column.OrderIndex = OrderIndex.Value;

            // The column nullability
            if (IsNotNull.HasValue && IsNotNull != column.IsNotNull)
                column.IsNotNull = IsNotNull.Value;

            // The column default value
            if (DefaultValue != enumColumn.DefaultValue)
                enumColumn.DefaultValue = DefaultValue;

            // The enumeration column value checking
            if (HasValueChecking.HasValue && HasValueChecking != enumColumn.HasValueChecking)
                enumColumn.HasValueChecking = HasValueChecking.Value;
        }


        /// <summary>
        ///   Creates a data transfer object for the specified enumeration column.
        /// </summary>
        /// <param name="enumColumn">The specified enumeration column.</param>
        /// <exception cref="ArgumentNullException">Thrown when the value of <paramref name="enumColumn"/> is null.</exception>
        /// <returns>Returns the data transfer object for the specified enumeration column.</returns>
        public static EnumColumnDto CreateFrom(EnumColumn enumColumn)
        {
            if (enumColumn == null)
                throw new ArgumentNullException("enumColumn");

            return new EnumColumnDto
            {
                Name = enumColumn.Name,
                DisplayName = enumColumn.DisplayName,
                Description = enumColumn.Description,
                OrderIndex = enumColumn.OrderIndex,
                IsNotNull = enumColumn.IsNotNull,
                DefaultValue = enumColumn.DefaultValue,
                HasValueChecking = enumColumn.HasValueChecking
            };
        }
    }
}
