﻿// -----------------------------------------------------------------------
// <copyright file="RefTargetDto.cs" company="JSC Centido">
//  Copyright (c) JSC Centido. All rights reserved.
// </copyright>
// -----------------------------------------------------------------------

using System;
using System.Globalization;
using System.Xml;
using System.Xml.Schema;

namespace Centido.Core.Dto
{
    /// <summary>
    ///   A data transfer object for a reference column target.
    /// </summary>
    public class RefTargetDto : DataTransferObject
    {
        /// <summary>
        ///   The name of the xml tag that is used to store this object.
        /// </summary>
        public const string XmlTag = "refTarget";

        /// <summary>
        ///   The current xml schema version.
        /// </summary>
        public const string CurrentXmlSchemaVersion = "1";

        /// <summary>
        ///   The name of the xml attribute that is used to store the need for cascade delete.
        /// </summary>
        public const string HasCascadeDeleteAttribute = "hasCascadeDelete";

        /// <summary>
        ///   The name of the xml attribute that is used to store the referenced table name.
        /// </summary>
        public const string RefTableNameAttribute = "refTableName";


        /// <summary>
        ///   Gets or sets the need for cascade delete.
        /// </summary>
        public bool? HasCascadeDelete { get; set; }


        /// <summary>
        ///   Gets or sets the referenced table name.
        /// </summary>
        public string RefTableName { get; set; }



        /// <summary>
        ///   This method is reserved and should not be used. When implementing the IXmlSerializable interface, you should return null (Nothing in Visual Basic) from this method, and instead, if specifying a custom schema is required, apply the <see cref="T:System.Xml.Serialization.XmlSchemaProviderAttribute"/> to the class.
        /// </summary>
        /// <returns>
        ///   An <see cref="T:System.Xml.Schema.XmlSchema"/> that describes the XML representation of the object that is produced by the <see cref="M:System.Xml.Serialization.IXmlSerializable.WriteXml(System.Xml.XmlWriter)"/> method and consumed by the <see cref="M:System.Xml.Serialization.IXmlSerializable.ReadXml(System.Xml.XmlReader)"/> method.
        /// </returns>
        public override XmlSchema GetSchema()
        {
            return null;
        }


        /// <summary>
        ///   Generates an object from its XML representation.
        /// </summary>
        /// <param name="reader">The <see cref="T:System.Xml.XmlReader"/> stream from which the object is deserialized. 
        /// </param>
        /// <exception cref="ArgumentNullException">Thrown when the value of <paramref name="reader"/> is null.</exception>
        public override void ReadXml(XmlReader reader)
        {
            if (reader == null)
                throw new ArgumentNullException("reader");

            reader.MoveToContent();

            // Reading the xml schema version
            var xmlSchemaVersion = ReadStringAttribute(reader, XmlSchemaVersionTag, isRequired: false);

            // If the xml schema version is not defined
            if (string.IsNullOrEmpty(xmlSchemaVersion)
                // or if it is the current xml schema version
                || xmlSchemaVersion == CurrentXmlSchemaVersion)
            {
                // Reading the references table name
                RefTableName = ReadStringAttribute(reader, RefTableNameAttribute);

                // Reading the cascade delete
                HasCascadeDelete = ReadBoolAttribute(reader, HasCascadeDeleteAttribute, isRequired: false);

                var isEmptyElement = reader.IsEmptyElement;

                reader.ReadStartElement();

                if (!isEmptyElement) reader.ReadEndElement();
            }
            else
                throw new ArgumentException(string.Format(
                    "The specified xml schema version is not supported (node = {0}, version = {1}).",
                    reader.Name,
                    xmlSchemaVersion));
        }


        /// <summary>
        ///   Converts an object into its XML representation.
        /// </summary>
        /// <param name="writer">The <see cref="T:System.Xml.XmlWriter"/> stream to which the object is serialized. 
        /// </param>
        /// <exception cref="ArgumentNullException">Thrown when the value of <paramref name="writer"/> is null.</exception>
        public override void WriteXml(XmlWriter writer)
        {
            if (writer == null)
                throw new ArgumentNullException("writer");

            writer.WriteStartElement(XmlTag);

            // The xml schema version
            writer.WriteAttributeString(XmlSchemaVersionTag, CurrentXmlSchemaVersion);

            // Writing the referenced table name
            writer.WriteAttributeString(RefTableNameAttribute, RefTableName);

            // Writing the cascade delete
            if (HasCascadeDelete.HasValue)
                writer.WriteAttributeString(HasCascadeDeleteAttribute, 
                    HasCascadeDelete.Value.ToString(CultureInfo.InvariantCulture));

            writer.WriteEndElement();
        }


        /// <summary>
        ///   Copies the properties of the data transfer object to the
        /// specified reference target.
        /// </summary>
        /// <param name="refTarget">The reference target.</param>
        /// <exception cref="ArgumentNullException">Thrown when the value of <paramref name="refTarget"/> is null.</exception>
        public void CopyTo(RefTarget refTarget)
        {
            if (refTarget == null)
                throw new ArgumentNullException("refTarget");

            if (HasCascadeDelete.HasValue && HasCascadeDelete != refTarget.HasCascadeDelete)
                refTarget.HasCascadeDelete = HasCascadeDelete.Value;
        }


        /// <summary>
        ///   Creates a data transfer object for the specified reference target.
        /// </summary>
        /// <param name="refTarget">The specified reference target.</param>
        /// <exception cref="ArgumentNullException">Thrown when the value of <paramref name="refTarget"/> is null.</exception>
        /// <returns>Returns the data transfer object for the specified reference target.</returns>
        public static RefTargetDto CreateFrom(RefTarget refTarget)
        {
            if (refTarget == null)
                throw new ArgumentNullException("refTarget");

            return new RefTargetDto
            {
                RefTableName = refTarget.Domain.GetTable(refTarget.RefTable.Id).Name,
                HasCascadeDelete = refTarget.HasCascadeDelete
            };
        }
    }
}
