﻿using System;
using System.Linq;

namespace Centido.Management.ConsoleManager
{
    /// <summary>
    ///   A console command that lists all security roles in the current domain.
    /// </summary>
    internal class ListRolesCommand : Command
    {
        /// <summary>
        ///   The display all option parameter.
        /// </summary>
        public const string DisplayAllOptionParameter = "all";

        /// <summary>
        ///   The display all option parameter (short name).
        /// </summary>
        public const string DisplayAllOptionShortParameter = "i";


        /// <summary>
        ///   Gets the console command name.
        /// </summary>
        public override string Name
        {
            get { return "listroles"; }
        }


        /// <summary>
        ///   Gets the console command description.
        /// </summary>
        public override string Description
        {
            get { return Messages.ListRolesCommandDescription; }
        }

        
        /// <summary>
        ///   Gets the console command full description.
        /// </summary>
        public override string FullDescription
        {
            get { return Messages.ListRolesCommandFullDescription; }
        }


        /// <summary>
        ///   Executes the console command.
        /// </summary>
        /// <param name="context">The console command execution context.</param>
        /// <exception cref="ArgumentNullException">Thrown when the value of <paramref name="context"/> is null.</exception>
        public override void Execute(CommandExecutionContext context)
        {
            var domain = context.CurrentDomain;

            if (domain == null) return;

            // If the current user has no permission to read the role list
            if (!domain.CurrentUser.CanChangeSecurity)
            {
                Console.WriteLine(Messages.NeedCanChangeSecurityPermissionMessage);
                return;
            }

            // Enumerating all roles
            foreach (var role in domain.Roles.OrderBy(r => r.Name))
            {
                // If the display all option is set
                if (context.GetParameterByName(DisplayAllOptionParameter, DisplayAllOptionShortParameter) != null)
                {
                    // If the display name is set
                    if (!string.IsNullOrEmpty(role.DisplayName))
                        Console.WriteLine("{0} ({1})", role.Name, role.DisplayName);
                    else
                        Console.WriteLine(role.Name);
                }
                else
                    Console.WriteLine(role.Name);
            }

            Console.WriteLine();
        }
    }
}
